:root {
  --font-color: #e5e0d7;
  --bg-color: #223159;
  --box-bg-color: transparent;
  --emph-color: #fe0052;
  --font-size: 18px;
  --content-width: 60ch;
  --narrow-content-width: 48ch;
  --content-padding: 1ch;
  --box-padding: 8px;
  --border-radius: 10px;
  --outline-width: 2px;
}

body {
  background-color: var(--bg-color); 
  color: var(--font-color);
  margin: 0;
  font-size: var(--font-size);
  font-family: Sans-Serif;
  height: calc(100vh - 4rem);
  text-align: center;
}
.content {
  min-height: calc(100vh - 4rem);
  max-width: var(--content-width);
  margin: auto;
  padding: var(--content-padding);
  vertical-align: middle;
  display: flex;
  flex-direction: column;
  justify-content: center;
  align-items: center;
}
footer {
  font-size: 75%;
  height: 2rem;
}
h1, h2, h3, h4, h5, h6, a, .explainer { color: var(--emph-color); }
h1, h2, h3, h4, h5, h6 { margin-bottom: 0.5rem; }
img + h1, svg + h1 { margin-top: 0; }
.narrow {
  max-width: min(
    var(--narrow-content-width),
    calc(100vw - calc(2 * var(--content-padding)))
  );
  margin: 0;
  padding: 0;
}

/*======================================================================
USER INPUT/OUTPUT
======================================================================*/

button {
  background-color: var(--emph-color);
  color: var(--font-color);
  font-size: 125%;
  font-weight: bold;
  padding: var(--box-padding);
  border: var(--outline-width) solid var(--emph-color);
  border-radius: var(--border-radius);
  text-align: center;
  cursor: pointer;
}
button:hover, button:focus {
  background-color: var(--font-color);
  color: var(--bg-color);
  border-color: var(--font-color);
}
input, textarea, .output-box {
  background-color: var(--box-bg-color);
  color: var(--font-color);
  text-align: left;
  border: var(--outline-width) solid var(--emph-color);
  border-radius: var(--border-radius);
  padding: var(--box-padding);
  resize: none;
}
.output-box {
  width: 100%;
  margin-top: 0.25em;
  margin-bottom: 0.25em;
}
textarea, form { width: 100% !important; }
textarea {
  scrollbar-width: none;
  -ms-overflow-style: none;
}
textarea::-webkit-scrollbar { display: none; }
input:focus, textarea:focus, .output-box {
  outline: none;
  border-color: var(--font-color);
}
.searchbar {
  display: flex;
  flex-wrap: nowrap;
  box-sizing: border-box;
  max-width: 100% !important;
  overflow: hidden;
}
.searchbar * {
  padding: var(--box-padding);
  flex-basis: 0;
}
.searchbar button {
  border-radius: 0 var(--border-radius) var(--border-radius) 0;
}
.searchbar input {
  font-size: 125%;
  min-width: 0;
  flex-grow: 1;
  border-radius: var(--border-radius) 0 0 var(--border-radius);
}

/*======================================================================
TABLES
======================================================================*/
.table-wrapper {
  overflow-x: auto;
  width: 100%;
}
table {
  min-width: 100%;
  table-layout: fixed;
  overflow-x: auto;
  border-collapse: separate;
  border-spacing: 0;
}
th, td { overflow: auto; }
td {
  border: solid calc(var(--outline-width) / 2) var(--font-color);
  padding: var(--box-padding);
  background-color: var(--box-bg-color);
}
td:first-child { border-left-width: var(--outline-width); }
td:last-child { border-right-width: var(--outline-width); }
tr:first-child td { border-top-width: var(--outline-width); }
tr:last-child td { border-bottom-width: var(--outline-width); }
tr:first-child td:first-child {
  border-top-left-radius: var(--border-radius);
  -webkit-border-top-left-radius: var(--border-radius);
  -moz-border-radius-topleft: var(--border-radius);
}
tr:first-child td:last-child {
  border-top-right-radius: var(--border-radius);
  -webkit-border-top-right-radius: var(--border-radius);
  -moz-border-radius-topright: var(--border-radius);
}
tr:last-child td:first-child {
  border-bottom-left-radius: var(--border-radius);
  -webkit-border-bottom-left-radius: var(--border-radius);
  -moz-border-radius-bottomleft: var(--border-radius);
}
tr:last-child td:last-child {
  border-bottom-right-radius: var(--border-radius);
  -webkit-border-bottom-right-radius: var(--border-radius);
  -moz-border-radius-bottomright: var(--border-radius);
}
